#!/usr/bin/env python

import gtk, gobject

# Options-system (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
# INFO:
# - a dynamic Options-system that allows very easy creation of
#   objects with embedded configuration-system.
#   NOTE: The Dialog is not very nice yet - it is not good OOP-practice
#   because too big functions and bad class-layout ... but it works
#   for now ... :)
#
# TODO:
# - cleanup __init__-handlers of Options and use **keyword_args instead
# - allow options to disable/enable other options
# - OptionGroup-class instead of (or behind) add_options_group
# - TimeOption, DateOption
# - FileOption needs filter-attribute
# - support for list-types (with popup like in gconf-editor)
# - support for EditableOptions-subclasses as options
# - separate OptionEditorWidget from Editor-Dialog


# -----------------------------------------------------------------------
# Option-classes and subclasses
# -----------------------------------------------------------------------

class Option(gobject.GObject):
	"""An Option stores information about a certain object-attribute.
It doesn't carry information about the value or the object it belongs to - 
it is only a one-way data-storage for describing how to handle attributes."""
	
	__gsignals__ = dict(option_changed=(gobject.SIGNAL_RUN_FIRST,
		gobject.TYPE_NONE, (gobject.TYPE_OBJECT,)))
	
	def __init__(self, group, name, default, label, desc,  
		disabled=False, hidden=False, callback=None):
		"""Creates a new Option with the given information."""
		super(Option, self).__init__()
		self.name = name
		self.label = label
		self.desc = desc
		self.default = default
		self.disabled = disabled
		self.hidden = hidden
		# for groups (TODO: OptionGroup)
		self.group= group
		# callback to be notified when this option changes
		self.callback = callback
		# real-time update?
		self.realtime = True
		
	def on_import(self, strvalue):
		"""Callback - called when an option gets imported from a string. 
This function MUST return the string-value converted to the required type!"""
		return strvalue.replace("\\n", "\n")
	
	def on_export(self, value):
		"""Callback - called when an option gets exported to a string. The
value-argument needs to be converted to a string that can be imported by the 
on_import-handler. This handler MUST return the value converted to a string!"""
		return str(value).replace("\n", "\\n")
	

class FileOption(Option):
	"""An Option-subclass for string-values that contain filenames. In fact 
nothing but a simple StringOption."""


class BoolOption(Option):
	"""An Option for boolean values."""
	
	def on_import(self, strvalue):
		if strvalue == "True":
			return True
		return False


class StringOption(Option):
	"""An Option for values of type string."""
	
	def __init__(self, group, name, default, label, desc, 
		disabled=False, hidden=False, callback=None, choices=None, ):
		Option.__init__(self, group, name, default,label, desc, 
		disabled=disabled, hidden=hidden, callback=callback)
		self.choices = choices


class IntOption(Option):
	"""An Option for values of type number (can be int or float)."""
	
	def __init__(self, group, name, default, label, desc,
		disabled=False, hidden=False, callback=None, 
		min=None, max=None, increment=1):
		Option.__init__(self, group, name, default, label, desc, 
			disabled=disabled, hidden=hidden, callback=callback)
		self.min = min
		self.max = max
		self.increment = increment
	
	def on_import(self, strvalue):
		"""Called when IntOption gets imported. Converts str to int."""
		if strvalue[0]=='-':
			return int(strvalue[1:]) * -1
		return int(strvalue)


class FloatOption(IntOption):
	"""An Option for values of type float."""
	
	def __init__(self, group, name, default, label, desc,
		disabled=False, hidden=False, callback=None, 
		min=None, max=None, increment=1, digits=1):
		IntOption.__init__(self, group, name, default, label, desc, 
			disabled=disabled, hidden=hidden, callback=callback,
			min=min, max=max, increment=increment)
		self.digits = digits
	
	def on_import(self, strvalue):
		"""Called when FloatOption gets imported. Converts str to float."""
		if strvalue[0]=='-':
			return float(strvalue[1:]) * -1.0
		return float(strvalue)


class FontOption(Option):
	"""An Option for fonts (a simple StringOption)."""


class ColorOption(Option):
	"""An Option for colors. Stored as a list with 4 values (r, g, b, a)."""
	
	def on_import(self, strvalue):
		"""Import (r, g, b, a) from comma-separated string."""
		tmpval = strvalue.split(',')
		outval = []
		for f in tmpval:
			outval.append(float(f))
		return outval
	
	def on_export(self, value):
		"""Export r, g, b, a to comma-separated string."""
		l = len(value)
		outval = ''
		for i in xrange(l):
			outval += str(value[i])
			if i<l-1:
				outval += ','
		return outval


# -----------------------------------------------------------------------
# EditableOptions-class
# -----------------------------------------------------------------------

class EditableOptions:
	"""The EditableOptions can be inherited from to allow objects to export 
editable options for editing them with the OptionsEditor-class:"""
	
	def __init__(self):
		self.__options__ = []
		self.__options_groups__ = {}
		# This is a workaround to remember the order of groups
		self.__options_groups_ordered__ = []
	
	def add_option(self, option, callback=None, realtime=True):
		"""Add an editable option to this object. Editable Options can """
		"""be edited and configured using the OptionsDialog. The optional """
		"""callback-arg can be used to set a callback that gets notified """
		"""when the option changes its value."""
		#print "Add option: "+option.name
		# if option already editable (i.e. initialized), return
		for o in self.__options__:
			if o.name == option.name:
				return False
		self.__dict__[option.name] = option.default
		# set auto-update (TEMPORARY?)
		option.realtime = realtime
		# add option to group (output error if group is undefined)
		try:
			self.__options_groups__[option.group]['options'].append(option)
		except:
			print "Options: Error- group "+option.group+" not defined."
			return False
		# now add the option
		self.__options__.append(option)
		# if callback is set, add callback
		if callback:
			option.connect("option_changed", callback)
		return True
		
		
	def add_options_group(self, name, group_info):
		"""Add a new options-group to this Options-object"""
		self.__options_groups__[name] = {'label':name, 
			'info':group_info, 'options':[]}
		self.__options_groups_ordered__.append(name)
		#print self.options_groups
	
	def disable_option(self, name):
		"""Disable the inputs for a certain Option."""
		for o in self.__options__:
			if o.name == name:
				o.disabled = True
				return True
		return False
	
	def export_options_as_list(self):
		"""Returns all editable options within a list (without groups)"""
		"""as key/value tuples."""
		lst = []
		for o in self.__options__:
			lst.append((o.name, getattr(self, o.name)))
		return lst
	
	def get_option_by_name(self, name):
		"""Returns an option in this Options by it's name (or None)."""
		"""TODO: this gives wrong results in childclasses ..."""
		for o in self.__options__:
			if o.name == name:
				return o
		return None
	
	def remove_option(self, name):
		"""Remove an option from this Options."""
		for o in self.__options__:
			if o.name == name:
				del o
				return True
		return True


# -----------------------------------------------------------------------
# OptionsDialog and UI-classes
# -----------------------------------------------------------------------

class OptionsDialog(gtk.Dialog):
	"""A first approach for creating a dynamic options-editor"""
	
	__shown_object = None
	#__option_groups = []
	#default_group_name = 'General'
	
	def __init__(self, width, height, treeview=False):
		# call gtk.Dialog.__init__
		super(OptionsDialog, self).__init__(
			"Edit Options", flags=gtk.DIALOG_DESTROY_WITH_PARENT |
			gtk.DIALOG_NO_SEPARATOR,
			buttons = (gtk.STOCK_CLOSE, gtk.RESPONSE_OK))
		# set size
		self.resize(width, height)
		self.set_keep_above(True)	# to avoid confusion
		# create header
		self.infohbox = gtk.HBox()
		self.infotext = None
		self.infoicon = None
		self.vbox.pack_start(self.infohbox, 0, -1, 10)
		sep = gtk.HSeparator()
		sep.show()
		#self.vbox.pack_start(sep, 0, -1)
		#self.vbox.add(self.infohbox)
		# add outer hbox
		self.outerhbox = gtk.HBox()
		self.outerhbox.show()
		#self.vbox.pack_start(self.outerhbox, 1, 1)
		print self.get_icon()
		self.vbox.add(self.outerhbox)
		# add listview on the left (maybe remove this?)
		if treeview:
			self.treeview = gtk.TreeView()
			self.treeview.set_size_request(50, 25)
			self.outerhbox.add(self.treeview)
			self.treeview.show()
		# create vbox for editor
		self.editorvbox = gtk.VBox()
		self.editorvbox.set_border_width(5)
		self.editorvbox.set_spacing(3)
		self.editorvbox.show()
		self.outerhbox.add(self.editorvbox)
		# crete tooltips-object
		self.tooltips = gtk.Tooltips()
	
	def set_info(self, name, info, copyright='', version='', icon=None):
		"""Add a header-information to this dialog."""
		# remove old entries
		if self.infotext:
			self.infotext.destroy()
			self.infoicon.destroy()
		# create infotext
		self.infotext = gtk.Label()
		self.infotext.use_markup = True
		markup = '<b><span size="xx-large">' + name + '</span></b>'
		if version:
			markup += '  <span size="large"><b>' + version + '</b></span>'
		markup += '\n'+info+'\n<span size="x-small">\n'+copyright+'</span>'
		self.infotext.set_markup(markup)
		self.infotext.set_line_wrap(True)
		self.infotext.set_alignment(0.0, 0.0)
		self.infotext.show()
		#self.infotext.set_size_request(self.width - 60, -1)
		# show icon?
		if icon:
			self.infoicon = icon
			self.infoicon.show()
			self.infohbox.pack_start(self.infoicon, 0, 1, 10)
		self.infohbox.pack_start(self.infotext, 1, 1, 5)
		#self.infohbox.set_homogeneous(False)
		# sow infobox
		self.infohbox.show()
		white = gtk.gdk.Color(1, 1, 1, 1)
		self.infohbox.modify_bg(gtk.STATE_NORMAL, white)
		self.infohbox.modify_base(gtk.STATE_NORMAL, white)
		

	#def add_editable_element(self, element):
	#	"""TODO: Add a new editable object to the list."""
	#	item = gtk.ListItem()
	#	self.treeview.add(item)
		
	def get_widget_for_option(self, option, value=None):
		"""Return a gtk.*Widget with Label within a HBox for a given option."""
		t = option.__class__
		widget = None
		if t == BoolOption:
			widget = gtk.CheckButton()
			widget.set_active(value)
			widget.connect("toggled", self.options_callback, option)
		elif t == StringOption:
			if option.choices:
				# if a list of values is defined, show combobox
				widget = gtk.combo_box_new_text()
				p = -1
				i = 0
				for s in option.choices:
					widget.append_text(s)
					if s==value:
						p = i
					i+=1
				widget.set_active(p)
				#widget.connect("changed", self.options_callback, option)
			else:
				widget = gtk.Entry()
				widget.set_text(value)
			#widget.connect("key-press-event", self.options_callback, option)
			widget.connect("changed", self.options_callback, option)
			#widget.set_size_request(180, 28)
		elif t == IntOption or t == FloatOption:
			widget = gtk.SpinButton()
			#widget.set_size_request(50, 22)
			#widget.set_text(str(value))
			if  t == FloatOption:
				widget.set_digits(1)
				widget.set_increments(0.1, int(option.max/100))
			else:
				widget.set_increments(1, int(option.max/100))
			widget.set_range(float(option.min), float(option.max))
			widget.set_value(value)
			widget.connect("value-changed", self.options_callback, option)
		elif t == ColorOption:
			widget = gtk.ColorButton()
			widget.set_use_alpha(True)
			widget.set_alpha(int(value[3]*65535))
			#col = gtk.gdk.Color(value[0], value[1], value[2])
			widget.set_color(gtk.gdk.Color(int(value[0]*65535), 
				int(value[1]*65535), int(value[2]*65535)))
			widget.connect("color-set", self.options_callback, option)
		elif t == FontOption:
			widget = gtk.FontButton()
			widget.set_font_name(value)
			widget.connect("font-set", self.options_callback, option)		
			"""elif t == OptionsType.LIST:
			widget = gtk.combo_box_new_text()
			for s in value:
				widget.append_text(s)
			widget.connect("changed", self.options_callback, option)"""
		elif t == FileOption:
			widget = gtk.FileChooserButton("Choose Icon")
			widget.set_filename(value)
			widget.set_size_request(180, 28)
			widget.connect("selection-changed", self.options_callback, option)			
		else:
			widget = gtk.Entry()
			print "unsupported type '"+str(t)+"'"
		hbox = gtk.HBox()
		label = gtk.Label()
		label.set_alignment(0.0, 0.0)
		label.set_label(option.label)
		label.set_size_request(180, 28)
		label.show()
		hbox.pack_start(label, 0, 1)
		if widget:
			if option.disabled:		# option disabled?
				widget.set_sensitive(False)
				label.set_sensitive(False)
			#label.set_mnemonic_widget(widget)
			self.tooltips.set_tip(widget, option.desc)
			widget.show()
			# check if needs Apply-button
			if option.realtime == False:
				but = gtk.Button('Apply', gtk.STOCK_APPLY)
				but.show()
				but.connect("clicked", self.apply_options_callback, 
					option, widget)
				b = gtk.HBox()
				b.show()
				b.pack_start(widget, 0, 0)
				b.pack_start(but, 0, 0)
				hbox.pack_start(b, 0, 0)
			else:
				#hbox.pack_start(widget, -1, 1)
				hbox.pack_start(widget, 0, 0)
		return hbox
	
	def reset_to_defaults(self):
		"""Reset all entries to their default values."""
		if self.__shown_object:
			for o in self.__shown_object.__options__:
				# set default value
				setattr(self.__shown_object, o.name, o.default)
	
	def read_option_from_widget(self, widget, option):
		"""Read an option's value from the widget and return it."""
		# get type of option and read the widget's value
		t = option.__class__
		if t == IntOption:
			val = int(widget.get_value())
		elif t == FloatOption:
			val = widget.get_value()
		elif t == StringOption:
			if option.choices:
				# if default is a list, handle combobox
				val = widget.get_active_text()
			else:
				val = widget.get_text()
		elif t == BoolOption:
			val = widget.get_active()
		elif t == ColorOption:
			col = widget.get_color()
			al = widget.get_alpha()
			val = (col.red/65535.0, col.green/65535.0, 
				col.blue/65535.0, al/65535.0)
		elif t == FontOption:
			val = widget.get_font_name()
		elif t == FileOption:
			val = widget.get_filename()
		else:
			print "OptionsDialog: Unknown option type: " + str(t)
			return None
		# return the value
		return val
	
	# TODO: custom callback/signal for each option?
	def options_callback(self, widget, optionobj):
		"""Callback for handling changed-events on entries."""
		#print "Changed: "+optionobj.name
		if self.__shown_object:
			# if the option is not real-time updated,
			if optionobj.realtime == False:
				return False
			# read option
			val = self.read_option_from_widget(widget, optionobj)
			if val != None:
				#print "SetOption: "+optionobj.name+"="+str(val)
				# set option
				setattr(self.__shown_object, optionobj.name, val)
				# notify option-object's on_changed-handler
				optionobj.emit("option_changed", optionobj)
		return False

	def apply_options_callback(self, widget, optionobj, entry):
		"""Callback for handling Apply-button presses."""
		if self.__shown_object:
			# read option
			val = self.read_option_from_widget(entry, optionobj)
			if val != None:
				#print "SetOption: "+optionobj.name+"="+str(val)
				# set option
				setattr(self.__shown_object, optionobj.name, val)
				# notify option-object's on_changed-handler
				optionobj.emit("option_changed", optionobj)
		return False

	def show_options_for_object(self, obj):
		"""Update the OptionsEditor to the options for the given Object."""
		self.__shown_object = obj
		# create notebook for groups
		notebook = gtk.Notebook()
		self.editorvbox.add(notebook)
		for group in obj.__options_groups_ordered__:
			group_data = obj.__options_groups__[group]
			# create box for tab-page
			page = gtk.VBox()
			if group_data['info'] != '':
				info = gtk.Label(group_data['info'])
				info.show()
				info.set_alignment(0, 0)
				page.pack_start(info, 0, 0, 7)
				sep = gtk.HSeparator()
				sep.show()
				#page.pack_start(sep, 0, 0, 5)
				
			# create VBox for inputs
			box = gtk.VBox()
			box.show()
			box.set_border_width(5)
			# add scrollwin
			scrollwin = gtk.ScrolledWindow()
			scrollwin.set_border_width(0)
			scrollwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
			scrollwin.add_with_viewport(box)
			scrollwin.set_shadow_type(gtk.SHADOW_NONE)
			scrollwin.show()
			page.add(scrollwin)
			page.show()
			# add new notebook-page
			label = gtk.Label(group_data['label'])
			label.show()
			notebook.append_page(page, label)
			# and create inputs
			for option in group_data['options']:
				if option.hidden == False:
					val = obj.__dict__[option.name]
					w = self.get_widget_for_option(option, val)
					if w:
						box.pack_start(w, 0, 0)
						w.show()
		notebook.show()

# ------ ONLY FOR TESTING ------------------:
if __name__ == "__main__":

	import os
	
	# this is only for testing - should be a Screenlet
	class TestObject(EditableOptions):

		def __init__(self):
			EditableOptions.__init__(self)
			# Add group
			self.add_options_group('General', 
				'The general options for this Object ...')
			self.add_options_group('Window', 
				'The Window-related options for this Object ...')
			self.add_options_group('Test', 'A Test-group ...')
			# Add editable options
			self.add_option(StringOption('Window', 'name', 'TESTNAME',
				'Testname', 'The name/id of this Screenlet-instance ...'), 
				realtime=False)
			self.add_option(StringOption('Window', 'name2', 'TESTNAME2',
				'String2', 'Another string-test ...'))
			self.add_option(StringOption('Test', 'combo_test', "el1", 'Combo', 
				'A StringOption displaying a drop-down-list with choices...', 
				choices=['el1', 'el2', 'element 3']), self.option_changed)
			self.add_option(FloatOption('General', 'flt', 30, 
				'A Float', 'Testing a FLOAT-type ...', 
				min=0, max=gtk.gdk.screen_width()))
			self.add_option(IntOption('General', 'x', 30, 
				'X-Position', 'The X-position of this Screenlet ...', 
				min=0, max=gtk.gdk.screen_width()))
			self.add_option(IntOption('General', 'y', 30, 
				'Y-Position', 'The Y-position of this Screenlet ...',
				min=0, max=gtk.gdk.screen_height()))
			self.add_option(IntOption('Test', 'width', 300, 
				'Width', 'The width of this Screenlet ...', min=100, max=1000))
			self.add_option(IntOption('Test', 'height', 150, 
				'Height', 'The height of this Screenlet ...', 
				min=100, max=1000))
			self.add_option(BoolOption('General', 'is_sticky', True, 
				'Stick to Desktop', 'Show this Screenlet always ...'))
			self.add_option(BoolOption('General', 'is_widget', False,  
				'Treat as Widget', 'Treat this Screenlet as a "Widget" ...'))
			self.add_option(FontOption('Test', 'font', 'Sans 14', 
				'Font', 'The font for whatever ...'))
			self.add_option(ColorOption('Test', 'color', (1, 0.35, 0.35, 0.7), 
				'Color', 'The color for whatever ...'))
			self.add_option(FileOption('Test', 'filename', os.environ['HOME'],
				'Filename-Test', 'Testing a FileOption-type ...'))
			# TEST
			self.disable_option('width')
			self.disable_option('height')
		
		#class TestObject2(EditableOptions):
			
		def option_changed(self, option, option2):
			"""Callback for handling when an option of this object changed."""
			"""Can be used to implement real-time saving to some backend."""
			print "OPTION CHANGED: "+option.name
			
		
	# this is only for testing - should be a Screenlet
	class TestObjectChild(TestObject):

		def __init__(self):
			TestObject.__init__(self)
			self.add_option(FileOption('Test', 'anothertest', 'ksjhsjgd',
				'Another Test', 'An attribute in the subclass  ...'))
	
	
	# TEST: load/save
	# TEST: option-editing
	to = TestObjectChild()
	#print to.export_options_as_list()
	se = OptionsDialog(450, 380)#, treeview=True)
	#img = gtk.image_new_from_stock(gtk.STOCK_ABOUT, 5)
	img = gtk.Image()
	img.set_from_file('notes/icon.svg')
	se.set_info('TestOptions', 
		'A test for an extended options-dialog with embedded about-info.' + 
		' Can be used for the Screenlets to have all in one ...\nNOTE:' + 
		'<span color="red"> ONLY A TEST!</span>',
		'(c) RYX 2007', version='v0.0.1', icon=img)
	se.show_options_for_object(to)
	resp = se.run()
	if resp == gtk.RESPONSE_OK:
		print "OK"
	else:
		print "Cancelled."
	se.destroy()
	
