#!/usr/bin/env python

#  NotesScreenlet (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
# INFO:
# - a simple sticky note application
# 
# TODO:
# - use clipping around cursor instead of full redraw
# - permanently save pango-layout instead of creating it on each draw
# - copy/paste
# - better handling of special-chars within text (on BACKSPACE)
# - maybe use another "paper" for the dragged-state
# - randomly rotate notes after they were "picked up"

import screenlets
from screenlets.options import IntOption, BoolOption, StringOption
from screenlets.options import FontOption, ColorOption
import gtk
import cairo
import pango
import sys
import random

class NotesScreenlet(screenlets.Screenlet):
	"""A sticky notes Screenlet with support for Themes."""
	
	# default meta-info for Screenlets
	__name__ = 'NotesScreenlet'
	__version__ = '0.4'
	__author__ = 'RYX (aka Rico Pfaus)'
	__desc__ = 'A sticky notes Screenlet with support for Themes and '+\
		'Pango-Markup within its text.'
	
	# internal vars
	__editing = True
	__cursor_index = 0
	__pin_var_rot = 0.0
	__pin_var_x = 0.0
	__pin_var_y = 0.0
	
	# editable options
	pin_x = 100
	pin_y = 6
	text_x = 19
	text_y =35
	font_name = 'Sans 12'
	rgba_color = (0.0, 0.0, 1.0, 1.0)
	text_prefix = '<b>'
	text_suffix = '</b>'
	random_pin_pos = True
	note_text = ""	# hidden option because val has its own editing-dialog
	
	# constructor
	def __init__(self, text="", **keyword_args):
		#call super
		screenlets.Screenlet.__init__(self, width=200, height=200, 
			uses_theme=True, **keyword_args) 
		# init attributes (not directly, that would cause redraw)
		self.__dict__['note_text'] = text
		self.__cursor_index = len(text)
		# set theme (redraws canvas)
		self.theme_name = "default"
		# add menu-items
		self.add_menuitem("edit_text", "Edit...")
		self.add_menuitem("", "-")
		self.add_menuitem("copy", "Copy")
		self.add_menuitem("paste", "Paste")
		self.add_menuitem("", "-")
		self.add_menuitem("clear", "Clear")
		# add default menuitems
		self.add_default_menuitems()
		# add key-handlers
		self.window.connect("key-press-event", self.key_press)
		# add settings groups
		self.add_options_group('Text', 
			'Text-/Font-related settings for the sticknotes.')
		self.add_options_group('Layout', 
			'The Layout-related settings for the sticknotes..')
		# add editable options
		self.add_option(IntOption('Layout', 'pin_x', 
			self.pin_x, 'X-Position of Pin', 
			'The X-Position of the tack/pin for the sticknote ...', 
			min=0, max=200))
		self.add_option(IntOption('Layout', 'pin_y', 
			self.pin_y, 'Y-Position of Pin', 
			'The Y-Position of the tack/pin for the sticknote ...', 
			min=0, max=200))
		self.add_option(BoolOption('Layout', 'random_pin_pos', 
			self.random_pin_pos, 'Randomize Pin', 
			'If active, the pin/tack will be slightly moved randomly '+\
			'whenever the note is picked up ...'))
		self.add_option(FontOption('Text', 'font_name', 
			self.font_name, 'Default Font', 
			'The default font of the text (when no Markup is used) ...'))
		self.add_option(ColorOption('Text', 'rgba_color', 
			self.rgba_color, 'Default Color', 
			'The default color of the text (when no Markup is used) ...'))
		self.add_option(IntOption('Layout', 'text_x', 
			self.text_x, 'X-Position of Text', 
			'The X-Position of the text-rectangle\'s upper left corner ...', 
			min=0, max=200))
		self.add_option(IntOption('Layout', 'text_y', 
			self.text_y, 'Y-Position of Text', 
			'The Y-Position of the text-rectangle\'s upper left corner ...', 
			min=0, max=200))
		self.add_option(StringOption('Text', 'text_prefix', 
			self.text_prefix, 'Text-Prefix', 
			'The text/Pango-Markup to be placed before the text ...'))
		self.add_option(StringOption('Text', 'text_suffix', 
			self.text_suffix, 'Text-Suffix', 
			'The text/Pango-Markup to be placed after the text ...'))
		# add hidden "note_text"-option (to save but not show in editor)
		self.add_option(StringOption('Text', 'note_text', 
			self.note_text, 'Note-Text', 
			'The text on this sticky note ...', hidden=True))
				
	def __setattr__(self, name, value):
		# call Screenlet.__setattr__ in baseclass (ESSENTIAL!!!!)
		screenlets.Screenlet.__setattr__(self, name, value)
		# check attribute name
		if name == "note_text":
			self.__dict__ [name]= value	# IMPORTANT: set value like this!!!
			self.redraw_canvas()
		elif name in ('width', 'height', 'pin_x', 'pin_y', 'text_x', 
			'text_y', 'text_suffix', 'text_prefix', 'random_pin_pos', 
			'font_name', 'rgba_color'):
			if self.window:
				self.redraw_canvas()
				self.update_shape()
	
	# TEMPORARY: should be replaced by real-time editing
	def show_edit_dialog(self):
		# create dialog
		dialog = gtk.Dialog("Edit Note", self.window)
		dialog.resize(self.width, self.height)
		dialog.add_buttons(gtk.STOCK_OK, gtk.RESPONSE_OK, 
			gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
		# create textview
		textview = gtk.TextView()
		textview.set_wrap_mode(gtk.WRAP_WORD)
		textbuffer = gtk.TextBuffer()
		textbuffer.set_text(self.note_text)
		textview.set_buffer(textbuffer)
		dialog.vbox.add(textview)
		textview.show()
		# run dialog
		response = dialog.run()
		if response == gtk.RESPONSE_OK:
			self.note_text = textbuffer.get_text(
				textbuffer.get_start_iter(), 
				textbuffer.get_end_iter())
		dialog.hide()
	
	def menuitem_callback(self, widget, id):
		screenlets.Screenlet.menuitem_callback(self, widget, id)
		if id=="edit_text":
			self.show_edit_dialog()
		elif id == 'copy':
			print "TODO: copy to clipboard"
		elif id == 'clear':
			self.note_text = ''
			self.__cursor_index = 0
		elif id == 'paste':
			print "TODO: paste text from clipboard"
	
	# EXPERIMENTAL!!!
	# TODO: cursor position
	def key_press(self, widget, event):
		"""Handle keypress events, needed for in-place editing."""
		print "Keyval: "+str(event.keyval)
		print "Key: "+str(event.string)
		code = event.keyval
		# HTML-special chars? &, <, >
		if code == 38:
			self.note_text += '&amp;'
		elif code == 60:
			self.note_text += '&gt;'
		elif code == 62:
			self.note_text += '&lt;'
		# valid ascii char?
		elif code > 20 and code < 256:
			self.__cursor_index += 1
			#TODO: add at cursor index
			self.note_text += event.string
			#self.redraw_canvas()
		# ENTER?
		elif code == 65293:
			self.__cursor_index += 1
			self.note_text += "\n"
		# BACKSPACE?
		elif code == 65288:
			if self.__cursor_index > 0:
				self.__cursor_index -= 1
				#TODO: remove at cursor index - 1
				self.note_text = self.note_text[:len(self.note_text)-1]
		# HOME? go to last EOL or pos -1
		elif code == 65360:
			p = self.__cursor_index
			while p >= 0 and self.note_text[p]!='\n':
				p -= 1
			if p>0:
				self.__cursor_index = p+1
			else:
				self.__cursor_index = -1
			self.redraw_canvas()
		# END?
		elif code == 65367:
			self.__cursor_index = len(self.note_text)-1
			self.redraw_canvas()
		# LEFT-ARROW?
		elif code == 65361:
			if self.__cursor_index > 0:
				self.__cursor_index -= 1
				self.redraw_canvas()
		# RIGHT-ARROW?
		elif code == 65363:
			# TODO: markup !!!!!!!!
			if self.__cursor_index < len(self.note_text):
				self.__cursor_index += 1
				self.redraw_canvas()
		print "cursorIndex: "+str(self.__cursor_index)

	def on_focus(self, event):
		if self.__editing == False:
			self.__editing = True
			self.redraw_canvas()
	
	def on_unfocus(self, event):
		if self.__editing == True and self.is_dragged == False:
			self.__editing = False
			self.redraw_canvas()
	
	def on_mouse_down(self, event):
		if event.button == 1:
			self.__pin_var_rot = (random.random()-0.5)/2
			self.__pin_var_x = random.random()
			self.__pin_var_y = random.random()
			self.redraw_canvas()
		return False

	def configure_event(self, widget, event):
		# if dragged before, redraw canvas
		if self.is_dragged==True:
			#print "Drag ended?"
			self.is_dragged=False
			self.redraw_canvas()
		# call super
		screenlets.Screenlet.configure_event(self, widget, event)
		return False
	
	""" TODO: make on_quit-handler work (return is ignored yet)
	def on_quit(self):
		dialog = gtk.MessageDialog(self.window, 
			gtk.DIALOG_MODAL, 
			gtk.MESSAGE_QUESTION, 
			gtk.BUTTONS_YES_NO)
		ret = False
		if dialog.run()==gtk.RESPONSE_NO:
			ret = True
		dialog.destroy()
		print "RETURN: "+str(ret)
		return ret
	"""
	
	# the drawing-handler, draws this Screenlet's background/visuals
	def on_draw(self, ctx):
		#set scale rel. to image size (the 200.0)
		#ctx.scale(self.width/200.0, self.height/200.0)
		ctx.scale(self.scale, self.scale)
		# translate while dragging
		if self.is_dragged:
			ctx.translate(-5, -5)
		#else:
			#else, add a small randomized rotation
			#ctx.rotate((random.random()-0.5)/25)
			# TODO: resize window to fit after rotation
		
		# render bg
		ctx.set_operator (cairo.OPERATOR_OVER)
		if self.theme.loaded:
			self.theme["note-bg.svg"].render_cairo(ctx)
		
		# get pango layout for self.window
		ctx.save()
		ctx.translate(self.text_x, self.text_y)
		#ctx.scale(200.0/self.width, 200.0/self.height)
		#ctx.scale(1,1)
		p_layout = ctx.create_layout()
		p_layout.set_wrap(pango.WRAP_CHAR)
		p_fdesc = pango.FontDescription(self.font_name)
		p_layout.set_font_description(p_fdesc)
		p_layout.set_width(((self.width - (self.text_x*2)) * pango.SCALE))
		p_layout.set_markup(self.text_prefix+self.note_text+self.text_suffix)
		ctx.set_source_rgba(self.rgba_color[0], 
			self.rgba_color[1], self.rgba_color[2], 
			self.rgba_color[3])
		ctx.show_layout(p_layout)
		ctx.fill()
		ctx.restore()
		
		# draw cursor
		if self.__editing:
			#print "draw cursor"
			if self.__cursor_index >= 0:
				cr = p_layout.index_to_pos(self.__cursor_index)
			else:
				cr = (0, 0, 2, 16)
			cr_w = 2
			cr_h = 16#cr[3]/pango.SCALE
			cr_x = self.text_x + (cr[0]/pango.SCALE)-1
			#if self.__cursor_index > -1:
			cr_x += cr[2]/pango.SCALE
			cr_y = self.text_y + (cr[1]/pango.SCALE)
			ctx.rectangle(cr_x, cr_y, cr_w, cr_h)
			#ctx.set_source_rgba(1, 1, 1, 0.7)
			ctx.fill()
			
		# draw pin if not dragging
		if self.is_dragged == False:
			ctx.translate(self.pin_x, self.pin_y)
			# add some variation?
			if self.random_pin_pos:
				ctx.rotate(self.__pin_var_rot)
				ctx.translate(self.__pin_var_x, self.__pin_var_y)
			# render pin
			ctx.set_operator (cairo.OPERATOR_OVER)
			if self.theme.loaded:
				self.theme["note-pin.svg"].render_cairo(ctx)
	
	# this handler draws the Screenlets window-shape to make
	# the window-background click-through in transparent areas
	def on_draw_shape(self,ctx):
		#set scale rel. to image size
		#ctx.scale(self.width/200.0, self.height/200.0)
		ctx.scale(self.scale, self.scale)
		# render bg
		if self.theme:
			self.theme["note-bg.svg"].render_cairo(ctx)


# If the program is run directly or passed as an argument to the python
# interpreter then create a Screenlet instance and show it
if __name__ == "__main__":
	text=""
	try:
		for i in range(1, len(sys.argv)):
			text += sys.argv[i] + " "
	except Exception, ex:
		print ex
		print "No text defined on commandline, using default."
		
	if text=="":
		text = """Sticknotes rule!"""
	#screenlet = NotesScreenlet(text)
	#screenlet.init_backend()
	#screenlet.main()
	sl = screenlets.create_new_instance('NotesScreenlet')
	sl.main()
	
