#!/usr/bin/env python

# Screenlets-Daemon (c) RYX (aka Rico Pfaus) 2007 <ryx@ryxperience.com>
#
# INFO:
#
# This is the Screenlets-daemon which cares for saving/restoring the
# current screenlets-session. It offers support for multiple backends 
# and is the recommended way to launch screenlets (if you want them all
# witihin one process and/or keeping their settings).
#
# NOTE: I didn't know how to pass objects over dbus so I use IDs
#       to identify instances - this MUST BE changed in the future!!!
#       It is pretty overcomplicated the way it works now ...
#
# dbus-iface methods:
# - IFace.register(classname): inform backend about newly opened 
#   screenlet and get a unique id for the new instance
# - IFace.save_option(id, name, value): to save an option
# - IFace.save_option_list(id, list): to save a list of options (as key/value)
# - IFace.add(classname): add a new Screenlet
# - IFace.unregister(id): remove screenlet and delete its config
#
# TODO:
# - support for restoring Screenlets as stand-alone apps while still
#   restoring their sessions
# - support for multiple sessions/setups (--session=<name> argument)
# - store state-files in Screenlets/.session/
# - ability to pass additional search-paths on the commandline to add to
#   the screenlets.SCREENLETS_PATH-list
#

import gobject
import random
import gtk
import sys
import os
import glob
import dbus
import dbus.service
if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
	import dbus.glib
from xdg import BaseDirectory

# import srceenlets (for module-loader and constants)
import screenlets

# the version of this backend
VERSION = '0.0.1'


class ScreenletsBackend:
	"""The backend performs the loading/saving of the 'key=value'-strings. 
Extend this superclass to implement different saving-backends."""
	
	def __init__(self):
		pass
	
	def delete_element(self, id):
		"""Delete an element's configuration by its id."""
		pass
	
	def save_option(self, id, name, value):
		"""Save one option for the element with the given id."""
		pass
	
	def load_option(self, id, name):
		"""Load one option for the element with the given id."""
		pass
	
	def load_element(self, id):
		"""Load all options for the element with the given id."""
		pass


class CachingBackend(ScreenletsBackend):
	"""A backend that stores the settings in arrays and saves after a short 
interval to avoid overhead when multiple values are set within a short time. 
The data gets saved into $HOME/.config/Screenlets/.session/, in a single file 
for each element (named like its id with the extension '.conf')."""
	
	# internals
	__elements = {}			# a dict with (id:dict)-entries cntaining the data
	__delay_time = 3000		# delay to wait before performing save
	__timeout = None		# the id of the timeout-function
	__queue = []			# list with ids of elements that need saving
	
	# attribs
	path = ''				# the path to store the files
	
	# Constructor
	def __init__(self, path):
		ScreenletsBackend.__init__(self)
		self.path = path
		self.__load_cache()
	
	def delete_element(self, id):
		"""Delete an element from the list and from the file-system."""
		if self.__elements.has_key(id):
			del self.__elements[id]
			try:
				import os
				os.remove(self.path + id + '.state')
			except Exception,ex:
				print ex
				print "Temporary file didn't exist - nothing to remove."
				return False
		print "CachingBackend: <#" + id + "> removed!"
		return True
		
	def save_option(self, id, name, value):
		"""Save option for an element to cache and start saving-timeout 
for that element (value must be of type string)."""
		# create key for option, if not existent yet
		if self.__elements.has_key(id) == False:
			self.__elements[id] = {}
		# set option in array
		self.__elements[id][name] = str(value)
		#print "Backend: "+name+"="+self.__elements[id][name]
		# if id is not already in queue, add the id to the queue
		if self.__queue.count(id) == 0:
			self.__queue.append(id)
		# reset timeout and start new
		if self.__timeout:
			gobject.source_remove(self.__timeout)
		self.__timeout = gobject.timeout_add(self.__delay_time, 
			self.__save_cache)#, id)
	
	def load_option(self, id, name):
		"""TODO: Load option from the backend (returned as str)."""
		return self.__elements[id][name]
	
	def load_element(self, id):
		"""Load all options for the element with the given id."""
		#print "Load element: "+id
		if self.__elements.has_key(id):
			return self.__elements[id]
		return None

	def __load_cache(self):
		"""Load all cached files from path."""
		# perform saving
		print "CachingBackend: Loading cache ..."
		# get dir content of self.path
		dirname = self.path
		dirlst = glob.glob(dirname + '*')
		tdlen = len(dirname)
		lst = []
		for fname in dirlst:
			dname = fname[tdlen:]
			if dname.endswith('.state'):
				id = dname[:-6]
				print "CachingBackend: Loading <"+id+">"
				#print "ID: "+id
				if self.__elements.has_key(id) == False:
					self.__elements[id] = {}
				# open file
				try:
					f = open(fname, 'r')
					lines = f.readlines()
					# read all options for this element from file
					for line in lines:
						#print "LOAD: "+line[:-1]
						parts = line[:-1].split('=', 1)
						if len(parts) > 1:
							self.__elements[id][parts[0]] = parts[1]
					f.close()
				except Exception, ex:
					print "Error while loading options: " + str(ex)
	
	def __save_cache(self):
		"""Save the cache (for all pending elements in queue) to self.path."""
		# loop through all elements in queue:
		for id in self.__queue:
			# if element with id not exists, remove it and break
			if self.__elements.has_key(id) == False:
				print "Queue-element <"+id+"> not found (already removed?)!"
				self.__queue.remove(id)
				break
			# create list with options
			print "CachingBackend: Saving <#"+id+"> :) ..."
			lst = []
			for oname in self.__elements[id]:
				lst.append([oname, self.__elements[id][oname]])
			# and save them (if any)
			if len(lst) > 0:
				try:
					f = open(self.path + id + '.state', 'w')
					for el in lst:
						f.write(el[0] + '=' + el[1] + "\n")
					f.close()
					print "OK"
				except:
					print "error while saving config: "+self.path+oname
		# clear queue
		self.__queue = []
		# NOT continue the timeout-function (!!!!!)
		return False

class ScreenletsService(dbus.service.Object):
	"""The ScreenletsService is a dynamic, dbus-based settings-daemon 
for the Screenlets. It cares for saving and restoring settings/options for 
Screenlets in real-time (the final saving/loading is done by the backend). 
Further it is the main thread for all Screenlets and used to keep all 
Screenlets within one process."""
	
	# constants
	ID_VALID_CHARS = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ' + \
		'abcdefghijklmnopqrstuvwxyz'
	ID_VALID_CHARS_NUM = len(ID_VALID_CHARS) - 1
	
	# only for notifications
	__NOTIFY_BUS = 'org.freedesktop.Notifications'
	__NOTIFY_OBJ = '/org/freedesktop/Notifications'
	__NOTIFY_IFACE =  'org.freedesktop.Notifications'
	
	# internals
	__screenlets = {}	# dict (id:classname) for all open Screenlets
	__next_id = 0		# TODO: should be read from config
	
	# settings (make args)
	restore_stand_alone = False
	
	# Constructor
	def __init__(self, bus_name, object_path=screenlets.BACKEND_OBJ):
		# call super
		dbus.service.Object.__init__(self, bus_name, object_path)
		# init libnotify
		self.notifications = dbus.Interface(dbus.SessionBus().get_object( \
			self.__NOTIFY_BUS, self.__NOTIFY_OBJ), self.__NOTIFY_IFACE)
		# init  backend (and create config path if not exists)
		cfg_dir = BaseDirectory.load_first_config('Screenlets')
		if cfg_dir == None:
			cfg_dir = BaseDirectory.save_config_path('Screenlets')
		if cfg_dir:
			cfg_dir += '/'
			self.backend = CachingBackend(path = cfg_dir)
		else:
			# no config-dir? use dummy-backend and note about problem
			self.backend = ScreenletsBackend()
			print "Unable to create config ... settings will not be saved!"
		# notify about self
		#self.notify('Screenlets-backend is running ...')
		# load session
		self.load_session()
		

	#--------------------------------------------------------------
	# public methods
	#--------------------------------------------------------------
	
	def create_instance_by_name(self, classname, is_widget=False):
		"""Creates a new Screenlet-instance by a classname and returns the 
added Screenlet-object."""
		print "create_instance_by_name: "+classname
		try:
			exec "from "+classname+" import "+classname
			classobj = eval(classname)
		except:
			return None
		if classobj:
			sl = classobj(is_widget = is_widget)
			if sl:
				#if is_widget:
				#	sl.is_widget = True
				return sl
			else:
				return None
	
	def create_id(self, digits=16):
		"""Create a (more or less) unique ID."""
		idstr = ''
		for i in xrange(digits):
			t1 = int(random.random() * self.ID_VALID_CHARS_NUM)
			idstr += self.ID_VALID_CHARS[t1]
		return idstr
		
	def get_next_id(self):
		"""Get a new unique id for a Screenlet."""
		#import uuid
		nid = self.create_id()
		tries = 0
		while self.__screenlets.has_key(nid):
			nid = self.create_id()
			tries += 1
			if tries > 1000:
				print "Error: Unable to get unique ID!"
				break
		return nid
		
	def get_screenlet_by_id(self, id):
		"""TODO: Return a Screenlet's data by its id."""
		# TODO: do we need this?
		pass
	
	def load_session(self):
		"""Load the session-file and restore all screenlets in it."""
		cfg_dir = BaseDirectory.load_first_config('Screenlets')
		if cfg_dir:
			filename = cfg_dir + '/screenletsd.ses'
			f=None
			try:
				f = open (filename, "r")
			except:
				print "Session-file " + str(filename) + " not found "+\
					"(will be created automatically)."
			if f:
				lines = f.readlines()
				for line in lines:
					# get name/id of Screenlet
					if len(line)>0:
						data = line.split(':', 1)
						id = data[0]
						if len(data)>0:
							classname = data[1].replace("\n", "")
							print "Restore: <"+classname+"#"+str(id)+">"
							# and restore the screenlet
							self.restore_instance(classname, id)
			return True
		else:
			print "No config found for screenletsd."
			pass
		return False
	
	def notify(self, message, timeout=2500):
		"""Show notification."""
		self.notifications.Notify('Screenlets-Daemon',
			0, '', 'Screenlets-Daemon', message, [], {}, timeout)
	
	def restore_instance(self, classname, id, stand_alone=False, 
		is_widget=False):
		"""Restore a (previously saved) Screenlet by classname/id."""
		# TEST: create as own process (maybe pass options as args?)
		if stand_alone:
			import os
			os.system('python -u ./'+classname+'.py &')
		else:
			#sl = self.create_instance_by_name(classname, is_widget=is_widget)
			# TEST: load module from path
			#sl = import_and_create_screenlet(classname)#, is_widget=is_widget)
			sl = screenlets.import_and_create_screenlet(classname)
			if sl:
				# not quit gtk when closed
				sl.quit_on_close = False
				# retrieve the Screenlet's options from backend
				opts = self.backend.load_element(id)
				if opts:
					for o in opts:
						# get Option-object for the attribute
						opt = sl.get_option_by_name(o)
						# NOTE: set attribute in Screenlet by calling the
						# on_import-function for the option (to import
						# the value as the required type)
						setattr(sl, opt.name, opt.on_import(opts[o]))
				else:
					print "No options found for "+classname
				# and add its id/classname to the list
				self.__screenlets[id] = classname
				# NOTE: Since the screenlet is launched within the
				# daemon's thread, the Screenlet.backend is replaced
				# by the daemon itself (because no need for dbus
				# if we are already within one application)
				# NOTE2: we do it AFTER setting the options to prevent
				# useless initial save-operations
				sl.backend = self
				sl.set_id(id)
				# and return the screenlet-object
				return sl
			else:
				print "Class "+classname+" not found."
				return False
	
	def save_session(self):
		"""Save the screenlets-session (called when Screenlets are added or 
removed. Stores ids and classnames to session-config."""
		cfg_dir = BaseDirectory.load_first_config('Screenlets')
		filename = cfg_dir + '/screenletsd.ses'
		f=None
		try:
			f = open (filename, "w")
		except Exception, ex:
			print "Exception: " + str(ex)
			print "screenletsd: Unable to save session."
		if f:
			print "Saving session to " + filename
			for s in self.__screenlets:
				f.write(s + ':' + self.__screenlets[s] + "\n")
			f.close()
			print str(len(self.__screenlets)) + ' Screenlet(s) opened in session.'
			return True
		return False
	
	#--------------------------------------------------------------
	# DBus-interface methods
	#--------------------------------------------------------------
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def add(self, classname, is_widget=False):
		"""Add a new Screenlet by its class-name."""
		print "ADD: "+classname
		if classname:
			sl = self.restore_instance(classname, str(self.get_next_id()), 
				is_widget=is_widget)
			if sl:
				self.save_session()
				return True
		return False
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def register(self, classname):
		"""Register an object with the backend. This function has to be
called by the Screenlet-superclass to allow saving/restoring options 
for a Screenlet. The function returns an id that further identifies the 
querying object within the daemon's class. Screenlets running within the 
daemon don't register themselves, instead they get registered when created."""
		# add new element
		id = str(self.get_next_id())
		self.__screenlets[id] = classname
		print 'Screenlet <'+classname + '_' + str(id)+'> has been registered!'
		# save session
		self.save_session()
		return id
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def save_option(self, id, name, value):
		"""Save an option for a Screenlet (i.e. pass it to the backend)."""
		self.backend.save_option(id, name, value)
		return True
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def save_option_list(self, id, lst):
		"""Save a 2d-list with options for a Screenlet. Can be used to set 
a larger amount of options in one call."""
		print "Received save-list request from <" + str(id) + ">"
		for entry in lst:
			self.backend.save_option(id, entry[0], entry[1])
		return True
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def unregister(self, id):
		"""Remove  Screenlet from the session."""
		print "Remove <"+str(id)+">"
		# if key is in list
		if self.__screenlets.has_key(id):
			# delete entry from list
			del self.__screenlets[id]
			# delete the Screenlet's from the backend
			self.backend.delete_element(id)
			print "Removed!"
		else:
			print "Not found in session."
		# save session
		self.save_session()
		return True
	
	@dbus.service.method(screenlets.BACKEND_IFACE)
	def test(self, obj):
		"""TEST."""
		print obj



#------------------------------------------------------------------------
# screenletsd - application code
#------------------------------------------------------------------------

# output info-message
#print 'Screenletsd v'+VERSION+' | (c) RYX (Rico Pfaus) 2007 | '+\
#	'released under the GPL'

# parse arguments
argc = len(sys.argv)
action = ''
if argc > 1:
	for i in range(1, argc):
		arg = sys.argv[i]
		if arg.startswith('--'):
			# check options
			print "Option "+arg+" not implemented"
	# first arg is action
	action = sys.argv[1]

# perform action
if action == '':
	print 'Usage: screenletsd.py <action> [options]'
	
elif action == 'start':
	# create new instance of ScreenletsService		
	session_bus = dbus.SessionBus()
	bus_name = dbus.service.BusName(screenlets.BACKEND_BUS, bus=session_bus)
	service = ScreenletsService(bus_name)
	# add sigkill-handler
	import signal
	def on_kill():
		global service
		service.notify('Screenlets-daemon has been killed .... ', 5000)
	signal.signal(signal.SIGTERM, on_kill)
	# start
	try:
		# start mainloop (gtk because we launch screenlets)
		gtk.main()
	except KeyboardInterrupt:
		# notify when daemon is closed
		service.notify('Screenlets-backend has been shut down .... ', 5000)
	except Exception, ex:
		print "Exception in ScreenletsService: " + ex

	del service

else:
	print "Error: unknown action '" + action + "'"
